# different work location
locations = {
    "rvk-office": {"desc": "Reykjavík office"},
    "hel-office": {"desc": "Helsinki office"},
    "tailscale": {"desc": "working over tailscale network, e.g. from home"},
    "external": {"desc": "working from the internet without VPN"},
}



import subprocess
import sys
import tkinter.messagebox
from typing import Optional

import click

from . import tools, ue

# our custom version of UnrealVersionSelector.exe

use_gui = sys.executable.endswith("pythonw.exe")


@click.group(invoke_without_command=False)
@click.option("--gui/--no-gui", default=use_gui, help="run with or without gui")
@click.pass_context
def uvs(ctx, gui):
    """A python version of the UnrealVersionSelector.exe.
    
    Initial command can be prefixed with a single dash or slash."""
    global use_gui
    use_gui = gui

    if ctx.invoked_subcommand is None:
        # if this were part of engine, we could register _this_ engine with prompt
        # but since it's not, we don't support this functionality
        # of invoking UnrealVersionSelector.exe with no arguments.
        pass
cli = uvs

@cli.command()
@click.option("--unattended", is_flag=True, help="Don't prompt for input")
@click.argument("engine_id", default=None)
def register(engine_id, unattended):
    """Register a _this_ engine.  Not supported."""
    click.echo("can't register _this_engine, because _this_ is not an engine")
    sys.exit(1)


# flags are non-standard extensions to the original UnrealVersionSelector.exe
@cli.command()
@click.option("--check", is_flag=True, help="Check current registrations.")
@click.option("--clear", is_flag=True, help="Remove associations.")
@click.option("--user", is_flag=True, help="use 'user' scope instead of 'machine'")
def fileassociations(check, clear, user):
    """Register file associations"""
    # TODO: call registry functions i ue module


# non-standard extension to the original UnrealVersionSelector.exe
@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def identifier(uproject):
    """Print the identifier for the project"""
    identifier = ue.get_identifier(uproject)
    click.echo(identifier)


# non-standard extension to the original UnrealVersionSelector.exe
@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def engineroot(uproject):
    """Print the engine root for the project"""
    root = ue.find_engine_for_project(uproject)
    click.echo(root)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
@click.option(
    "--engine_id", metavar="<engine_id>", default=None, help="The engine id to use"
)
def switchversion(uproject, engine_id):
    """Switch the engine version for the project"""
    # TODO, perform this interactive operation
    if engine_id:
        # switch_engine_id_silent(uproject, engine_id)
        pass
    else:
        # switch_engine_id_prompt(uproject)
        pass


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
@click.argument("engine_id", nargs=1, metavar="<engine_id>")
def switchversionsilent(uproject, engine_id):
    """Switch set the engine-id for this project to a given engine-id without prompting"""
    # TODO, perform this interactive operation
    # switch_engine_id_silent(uproject, engine_id)
    pass


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def editor(uproject):
    """Start the editor for project"""
    editor = ue.get_editor_from_uproject()
    if editor:
        result = subprocess.run([editor, uproject])
        sys.exit(result.returncode)
    else:
        click.echo("No editor found")
        sys.exit(1)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def projectlist(uproject):
    """Start the editor with a list of projects"""
    editor = ue.get_editor_from_uproject()
    if editor:
        result = subprocess.run([editor])
        sys.exit(result.returncode)
    else:
        click.echo("No editor found")
        sys.exit(1)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def game(uproject):
    """Start the editor for the project and run the game"""
    editor = ue.get_editor_from_uproject()
    if editor:
        result = subprocess.run([editor, uproject, "-game"])
        sys.exit(result.returncode)
    else:
        click.echo("No editor found")
        sys.exit(1)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def projectfiles(uproject):
    """Rebuild project files for the project"""
    # editor = ue.get_editor_from_uproject()
    pass


@cli.command()
@click.argument("uproject", type=click.Path(exists=False), metavar="<.uproject>")
def test(uproject):
    messagebox_ok("This is a test message", True)
    messagebox_ok("This is a test message", False)


@cli.command()
def register_test():
    """Register uvs as the default UnrealVersionSelector"""
    print(sys.argv)
    print(__spec__)
    return
    # find the command used to run us
    cmd = sys.argv[:]
    i = cmd.index("register-test")
    cmd = cmd[:i]
    # remove any trailing args
    while cmd[-1].startswith("-"):
        cmd.pop()

    # if we are running a .py file, we need to use the -m flag with the actual
    # module name (not __main__)
    if cmd[0].endswith(".py"):
        exe = sys.executable
        if exe.endswith("python.exe"):
            exe = exe[:-10] + "pythonw.exe"
        cmd = [exe, "-m", __spec__.name] + cmd[1:]

    ue.register_uproject_handler(engine_path=None, handler=cmd, test=True)


def main():
    # ignore initial double dash arguments though, which are
    # global
    i = 1
    while i < len(sys.argv) and sys.argv[i].startswith("--"):
        i += 1

    # now we get to the command.  strip a single / or - from it.
    if len(sys.argv) > i:
        # turn slash option into a command, stripping the slash
        if sys.argv[i].startswith("/"):
            sys.argv[i] = sys.argv[i][1:]
        elif sys.argv[i].startswith("-") and not sys.argv[i].startswith("--"):
            sys.argv[i] = sys.argv[i][1:]

    # also allow single dash options, by turning them into two dash options
    for i, arg in enumerate(sys.argv[1:]):
        if arg.startswith("-") and not arg.startswith("--"):
            sys.argv[i + 1] = f"--{arg[1:]}"

    tools.click_main(cli, obj={})


def messagebox_ok(
    message: str, success: bool = True, heading: Optional[str] = None
) -> bool:
    if not heading:
        heading = "Success" if success else "Error"

    if use_gui:
        if success:
            tkinter.messagebox.showinfo(heading, message)
        else:
            tkinter.messagebox.showerror(heading, message)
    else:
        # just use echo
        if success:
            click.echo(message)
        else:
            click.echo(
                click.style(f"{heading}:", fg="red") + f" {message}", file=sys.stderr
            )

    return success


if __name__ == "__main__":
    main()



from __future__ import annotations

import ntpath
import winreg
from typing import Any, Iterator, Optional, Tuple, Union

"""helper classes for registry operations"""

# map the winreg root key constants to their names
root_keys: dict[int, str] = {}
for key, val in winreg.__dict__.items():
    if key.startswith("HKEY_"):
        root_keys[val] = key
        globals()[key] = val


class Key:
    """A key in the registry.  This is a wrapper around the winreg module."""

    def __init__(
        self,
        parent: Union[Key, int],
        name: str,
        **kwargs,
    ) -> None:
        """Create a new key object.  The key is not opened or created."""
        self._parent = parent
        self.name: str = name
        self._handle: Optional[Any] = None
        self._kwargs = kwargs  # passed to an implicit open() in context manager
        if isinstance(parent, Key):
            self._fullname = ntpath.join("Computer", parent._fullname, name)
        else:
            self._fullname = ntpath.join(root_keys[parent], name)

    def __repr__(self) -> str:
        """Returns a string representation of the key"""
        return f"Key<{self._fullname!r}>"

    def _get_basename(self) -> Tuple[Any, str]:
        """Returns a suitable key handle and base string for this key"""
        if self._handle is not None:
            return self._handle, ""
        if isinstance(self._parent, Key):
            parent_handle, parent_name = self._parent._get_basename()
            return parent_handle, ntpath.join(parent_name, self.name)
        # parent is a raw handle
        return self._parent, self.name

    def open(
        self,
        *subkeys: str,
        create: bool = False,
        write: Optional[bool] = None,
        check: bool = True,
    ) -> Optional[Key]:
        """Opens an existing key"""
        if subkeys:
            return self.subkey(*subkeys).open(create=create, write=write, check=check)
        assert self._handle is None
        handle, name = self._get_basename()
        try:
            # create defaults for true if create is true, false otherwise
            write = write if write is not None else create
            access = winreg.KEY_WRITE if write else winreg.KEY_READ
            func = winreg.CreateKeyEx if create else winreg.OpenKeyEx
            self._handle = func(handle, name, access=access)
        except FileNotFoundError as e:
            if check:
                raise ValueError(f"Key {self.name!r} not found") from e
            return None
        return self

    def subkey(self, *subkeys: str, **kwargs) -> Key:
        """Returns a subkey object"""
        subkey_name = ntpath.join(*subkeys)
        return Key(self, subkey_name, **kwargs)

    def __getattr__(self, name: str) -> Key:
        """Returns a subkey object"""
        return self.subkey(name)

    def __call__(self, *subkeys: str, **kwargs) -> Key:
        """Returns a subkey object"""
        return self.subkey(*subkeys, **kwargs)

    def exists(self) -> bool:
        """checks if the key exists"""
        if self.opened():
            return True
        if self.open(check=False) is None:
            return False
        self.close()
        return True

    def __bool__(self) -> bool:
        """checks if the key exists"""
        return self.exists()

    def opened(self) -> bool:
        """Checks if the key is opened"""
        return self._handle is not None

    def create(self, *args, **kwargs) -> Key:
        """alias for open(... create=True)"""
        kwargs["create"] = True
        return self.open(*args, **kwargs)

    def close(self) -> None:
        """Closes the key."""
        if self._handle is not None:
            handle, self._handle = self._handle, None
            winreg.CloseKey(handle)

    # iterating over the key/value pairs (items) in the key, similar to a dict.

    def items(self) -> Iterator[tuple[str, tuple[Any, int]]]:
        """Iterates over the values in the key"""
        for i in range(1000):
            try:
                name, value, type = winreg.EnumValue(self._handle, i)
                yield (name, (value, type))
            except OSError:
                break

    def keys(self) -> Iterator[str]:
        """iterates of the item names in the key"""
        for itemname, _ in self.items():
            yield itemname

    def values(self) -> Iterator[tuple[Any, int]]:
        """iterates of the item values in the key"""
        for _, value in self.items():
            yield value

    def stringitems(self) -> Iterator[tuple[str, str]]:
        """Iterates over the string items in the key"""
        for k, (v, t) in self.items():
            if t == winreg.REG_SZ:
                yield (k, v)

    def subkeys(self) -> Iterator[Key]:
        """Iterates over the subkeys in the key"""
        for i in range(1000):
            try:
                name = winreg.EnumKey(self._handle, i)
                yield Key(self, name)
            except OSError:
                break

    def get_value(self, name: str, default: Any = None) -> Any:
        """Gets a value from the key."""
        try:
            return winreg.QueryValueEx(self._handle, name)[0]
        except FileNotFoundError:
            return default

    def get_value_ex(self, name: str, default: Any = None) -> tuple[Any, int]:
        """Gets a value from the key, returning the value and a type."""
        if not name:
            name = ""  # get the default value
        try:
            return winreg.QueryValueEx(self._handle, name)
        except FileNotFoundError:
            return default, None

    def set_value(self, name: str, value: Any, type=winreg.REG_SZ) -> None:
        """Sets a value in the key"""
        if not name:
            name = ""  # set the default value
        winreg.SetValueEx(self._handle, name, 0, type, value)

    def get(self, name: str, default: Any = None) -> Any:
        """Gets a value from the key"""
        return self.get_value(name, default)

    def __getitem__(self, name: str) -> Tuple[Any, int]:
        """Get a value from the key"""
        try:
            return winreg.QueryValueEx(self._handle, name)[0]
        except FileNotFoundError as e:
            raise KeyError(name) from e

    def __setitem__(self, name: str, value: str) -> None:
        """Sets a value in the key. We assume a string"""
        return self.set_value(name, value)

    def __delitem__(self, name: str) -> None:
        """Deletes a value from the key"""
        try:
            winreg.DeleteValue(self._handle, name)
        except FileNotFoundError as e:
            raise KeyError(name) from e

    def __enter__(self) -> Key:
        """Context manager: opens the key if required"""
        if not self.opened():
            return self.open(**self._kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def delete(self, tree: bool = False, missing_ok=True) -> None:
        """Deletes the key, optionally recursively."""
        if self.opened():
            raise ValueError("Cannot delete open key")
        if missing_ok and not self.exists():
            return
        if tree:
            with self.open():
                for subkey in list(self.subkeys()):
                    subkey.delete(tree=True)
        winreg.DeleteKey(self._parent, self.name)

    def copy(self) -> Key:
        """Returns a fresh, un-opened copy of the key"""
        return Key(self._parent, self.name)

    def print(self, tree=False, indent=4, level=0):
        """Prints the key to stdout"""
        print(" " * level * indent + f"key: '{self.name}'")
        with self.copy().open() as key:
            for name, value in key.items():
                print(" " * (level + 1) * indent + f"val: '{name}' = {value})")
            if not tree:
                for sub in key.subkeys():
                    print(" " * (level + 1) * indent + f"key: '{sub.name}'")
            else:
                for sub in key.subkeys():
                    sub.print(tree=True, indent=indent, level=level + 1)



"""Configuration for the mainframe_paxo module."""
import os.path

# folder on the work drive that we use
work_drive_dev_folder = os.path.join("/", "paxdei_dev")
work_drive_ddc_folder = os.path.join("/", "paxdei_DDC")

# the drive letter we use for the subst
# the letter P: has been taken to be used by Pipeline.  Let's use Q: instead.
subst_drive_name = "Q:"


# p4 fingerprints for the different servers
p4_fingerprints = {
    "aws": "7F:24:67:0B:62:7B:9F:3A:ED:5F:26:32:23:82:8F:20:EE:13:8B:03",
    "rvk": "8C:07:9D:29:F8:03:CC:76:C0:3B:26:41:20:3D:4C:B0:F0:A4:5E:B8",
    "rvk-old": "B9:62:8C:DA:75:B7:85:0E:B1:2B:02:1A:AE:11:5B:25:7D:C8:72:CF",
    "hel": "62:95:05:A0:B4:AD:1A:61:12:F6:06:07:9A:D3:5C:83:85:A6:67:C9",
}

# configurations for the different locations
locations = {
    "rvk-office": {
        "desc": "Reykjavík office",
        "p4port": "ssl:p4-rvk.mainframe.zone:1666",
        "p4trust": p4_fingerprints["rvk"],
        "ddc": r"\\ddc-rvk.mainframe.zone\DDC",
    },
    "rvk-ext": {
        "desc": "Reykjavík, accessin office from the internet",
        "p4port": "ssl:p4-rvk.x.mainframe.zone:1666",
        "p4trust": p4_fingerprints["rvk"],
        "ddc": None,
    },
    "hel-office": {
        "desc": "Helsinki office",
        "p4port": "ssl:p4-hel.mainframe.zone:1666",
        "p4trust": p4_fingerprints["hel"],
        "ddc": r"\\ddc-hel.mainframe.zone\DDC",
    },
    "hel-ext": {
        "desc": "Helsinki office",
        "p4port": "ssl:p4-hel.x.mainframe.zone:1666",
        "p4trust": p4_fingerprints["hel"],
        "ddc": None,
    },
    "tailscale": {
        "desc": "working over tailscale network, e.g. from home",
        "p4port": "ssl:p4.t.mainframe.zone:1666",
        "p4trust": p4_fingerprints["aws"],
        "ddc": r"\\ddc.t.mainframe.zone\DDC",
    },
    "external": {
        "desc": "working from the internet without VPN",
        "p4port": "ssl:perforce.x.mainframe.zone:1666",
        "p4trust": p4_fingerprints["aws"],
        "ddc": None,
    },
}



import subprocess

# stuff to install with winget


def install(id, force=False, upgrade=False):
    cmd = ["winget", "install", "-e", "--id", id]
    if force:
        cmd.append("--force")
    if upgrade:
        cmd[1] = "upgrade"
    subprocess.run(cmd)



import os.path
import uuid
from typing import Dict, Optional

from ..registry import HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE, Key
from . import desktop


def normalize_root(path: str) -> str:
    return os.path.normcase(os.path.normpath(path))


def enumerate_engine_installations() -> Dict[str, str]:
    launcher_engines = enumerate_launcher_engine_installations()
    user_engines = enumerate_user_engine_installations()
    launcher_engines.update(user_engines)
    return launcher_engines


def enumerate_launcher_engine_installations() -> Dict[str, str]:
    """Enumerate the engine installations from the launcher"""

    result = {}
    with Key(
        HKEY_LOCAL_MACHINE,
        "SOFTWARE\\EpicGames\\Unreal Engine\\Builds",
    ) as key:
        for subkey in key.subkeys():
            with subkey:
                name = subkey.name
                root = subkey.get("InstalledDirectory", None)
                if root and desktop.is_valid_root_directory(root):
                    result[name] = normalize_root(root)
    return result


def enumerate_user_engine_installations() -> Dict[str, str]:
    """Enumerate the engine installations from the launcher"""

    result = {}
    remove = []  # values to remove
    unique_directories = set()

    with Key(
        HKEY_CURRENT_USER, "SOFTWARE\\Epic Games\\Unreal Engine\\Builds", write=True
    ) as key:
        if not key:
            return result
        for k, v in key.items():
            v = normalize_root(v)
            ok = False
            if desktop.is_valid_root_directory(v):
                try:
                    guid = uuid.UUID(k)
                except ValueError:
                    guid = None
                if guid:
                    # clean out duplicate guid keys for the same engine
                    if v not in unique_directories:
                        unique_directories.add(v)
                        ok = True
                else:
                    # we accept all non-guid keys, since they were created by the user
                    ok = True
            if ok:
                result[k] = v
            else:
                remove.append(k)

        # Remove all the keys which weren't valid
        for k in remove:
            del key[k]
    return result


def register_engine_installation(
    root_dir: str, engine_id: Optional[str]
) -> Optional[str]:
    """Register an engine installation"""
    if not engine_id:
        engine_id = str(uuid.uuid4())
        is_guid = True
    else:
        is_guid = desktop.is_guid(engine_id)
    root_dir = normalize_root(root_dir)
    if not desktop.is_valid_root_directory(root_dir):
        return None

    with Key(
        HKEY_CURRENT_USER, "SOFTWARE\\Epic Games\\Unreal Engine\\Builds", create=True
    ) as key:
        if is_guid:
            # check if there is a guid key with the same root dir
            for k, v in key.items():
                if is_guid(k) and normalize_root(v) == root_dir:
                    # we just re-use the existing guid key
                    return k
        key[engine_id] = root_dir
    return engine_id



# utilities for perforce

import contextlib
import datetime
import getpass
import json
import os
import os.path
import socket
import subprocess
import xml.etree.ElementTree as ElementTree

import click

from . import tools, winget
from .config import locations

# path prefix for p4 command
p4_prefix = None

# the current user name
p4_user = None

# fingerprints for various p4 servers


def p4_run(args, **kwargs):
    """run a p4 command, and return the output"""

    return p4_run_raw(args, **kwargs)


def p4_json(args, check=True, **kwargs):
    """run a p4 command.  Input and output are dicts"""

    # add options for json input/output
    options = ["-ztag", "-Mj"]

    kwargs["capture_output"] = True

    out = p4_run_raw(options + args, check=check, **kwargs)
    return parse_json_output(out.stdout.decode())


def parse_json_output(data):
    # P4 returns lines of json.  Sometimes there is unparsable data, and we return that
    # as is.
    """json output from p4 is in a line, followed by extra data"""
    result = []
    for line in data.split("\n"):
        if not line:
            continue
        try:
            result.append(json.loads(line))
        except json.JSONDecodeError:
            result.append(line)
    return result


def p4_run_raw(cmd, **kwargs):
    # We have logic here to be able to find p4 even if we just installed it and
    # it is not in the PATH
    global p4_prefix
    if p4_prefix:
        p4 = os.path.join(p4_prefix, "p4.exe")
    else:
        p4 = "p4"

    # inherit and override environment variables
    if "env" in kwargs:
        kwargs["env"] = dict(os.environ, **kwargs["env"])
    try:
        # print(f"running {cmd} with {kwargs}")
        return subprocess.run([p4] + cmd, **kwargs)
    except FileNotFoundError:
        if p4_prefix:
            raise
        p4_prefix = os.path.join(os.environ["ProgramFiles"], "Perforce")
        p4 = os.path.join(p4_prefix, "p4.exe")
        return subprocess.run([p4] + cmd, **kwargs)


location_type = click.Choice(locations.keys(), case_sensitive=False)


def validate_work_drive(cxt, param, drive: str):
    if len(drive) == 1:
        drive = drive + ":"
    if len(drive) != 2 or drive[1] != ":" or not drive[0].isalpha():
        raise click.BadParameter(
            "work drive must be a single letter followed by a colon"
        )
    if drive.upper() == "P:" or not os.path.isdir(drive):
        raise click.BadParameter(f"{drive} is not a valid drive")
    return drive.upper()


@click.group(chain=True)
def p4():
    """Various Perforce tools for PaxDei development.
    For initial setup use:

    'paxo p4 install setup sync'."""
    pass


@p4.command()
@click.option("--upgrade", is_flag=True, help="upgrade p4 if already installed")
@click.option("--force", is_flag=True, help="force install/upgrade")
def install(force, upgrade):
    """Install/upgrade p4 and p4v"""
    p4_install(force=force, upgrade=upgrade)


@p4.command()
@click.option(
    "--location",
    type=location_type,
    prompt="Specify your location",
    required=True,
    default=lambda: tools.location_get(empty_ok=True),
)
@click.option(
    "--work-drive",
    type=str,
    required=True,
    prompt="Specify your work drive (e.g. D:)",
    callback=validate_work_drive,
    default=tools.workdrive_get(empty_ok=True),
)
@click.option(
    "--username",
    type=str,
    prompt="Perforce username",
    default=lambda: get_username(),
    help="specify the Perforce user name",
)
def setup(location, work_drive, username):
    """Set up Perforce workspaces for development"""
    p4_setup_dev(location, work_drive, username)


@p4.command()
@click.option("--depot", type=str, default="all")
def sync(depot):
    """sync the given depot"""
    all = depot == "all"
    if all:
        msg = "syncing all depots"
    else:
        msg = f"syncing depot '{depot}'"
    print(msg)
    p4_sync(depot, all)
    print(msg + "sync  done")


def p4_install(force=False, upgrade=False):
    """Initial setup of p4 for PaxDei development"""
    print("Installing P4 and P4V")

    # optionally install p4
    version = have_p4()
    if version:
        print(f"current p4 version {version}")
        if not upgrade or force:
            print("Not installing. use --upgrade to upgrade")
            return
    print("p4 not found")
    version = install_p4(force=force, upgrade=upgrade)
    print(f"p4 version {version}")

    # various settings
    p4_set_various()

    # fixup p4v settings
    fix_p4v_settings()


def install_p4(force=False, upgrade=False):
    version = have_p4()
    winget.install("Perforce.P4V", force=force, upgrade=upgrade)
    if not version:
        # we just installed it, and want it in the path
        tools.addpath(os.path.join(os.environ["ProgramFiles"], "Perforce"))
    return have_p4()


def p4_setup_dev(location, work_drive, username):
    # set p4port and p4trust
    tools.location_set(location)
    set_location(location)

    # create the subst
    tools.subst_drive(work_drive, force=True)

    p4_set_user(username)

    # set the client to the paxdei depot by default
    p4client_set(get_client_name("paxdei"))

    # login
    do_login(quiet=True)

    # create the client specs
    tools.workdrive_set(work_drive)
    for depot in ["paxdei", "UE"]:
        do_create_client(work_drive, depot)


@p4.command()
def get_version():
    """Get the version of the installed Perforce client."""
    version = have_p4()
    if version:
        print(f"p4 version {version}")
    else:
        print("p4 not found")


def do_login(quiet=False):
    username = p4_set_get()["P4USER"]
    out = p4_run(["login", "-s"], capture_output=True)
    if out.returncode == 0:
        if not quiet:
            print(
                f"User '{username}' already logged in. " + out.stdout.decode().strip()
            )
        return True

    print(
        f"Not logged in.  Attempting login for user '{username}'. Enter password and look for a JumpCloud notification on your phone."
    )
    out = p4_run(["login"])
    if out.returncode != 0:
        print("login failed.")
        print(
            "If the above message says that your account 'has not been enabled', please contact #it_helpdek"
        )
        print("to have your account enabled for Perforce access, before retrying.")
        raise click.ClickException("login failed")

    return out.returncode == 0


def set_location(location):
    p4port = get_p4port(location)
    print(f"setting P4PORT to {p4port}")
    p4_run(["set", f"P4PORT={p4port}"], check=True)
    p4trust = get_p4trust(location)
    p4_run(["trust", "-f", "-i", p4trust], check=True)


def get_p4port(location):
    """return a suitable value for P4PORT"""
    return locations[location]["p4port"]


def get_p4trust(location):
    """return a suitable value for P4TRUST"""
    loc = locations[location]
    return loc["p4trust"]


def have_p4():
    version = None
    try:
        output = p4_run(["-V"], capture_output=True, check=True)
        for line in output.stdout.decode().split("\n"):
            if line.startswith("Rev."):
                version = line.split()[1].split("/")[2]
                break
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass

    return version


def p4_set_various():
    """Set various p4 options"""
    p4_run(["set", "P4CHARSET=utf8"])
    p4_run(["set", "P4COMMANDCHARSET=utf8"])
    p4_run(["set", "P4CONFIG=.p4config"])
    p4_run(["set", "P4IGNORE=p4ignore.txt;.p4ignore"])


def p4_set_get():
    out = p4_run(["set"], capture_output=True, check=True)
    result = {}
    for line in out.stdout.decode().split("\n"):
        line = line.strip()
        if not line:
            continue
        # remove the trailing () comment
        line = line.rsplit(" ", 1)[0]
        key, value = line.split("=", 1)
        result[key] = value
    return result


def p4_set_user(user):
    """Set the user name"""
    global p4_user
    user = user or getpass.getuser()
    p4_run(["set", f"P4USER={user}"])
    p4_user = user


@p4.command()
@click.option(
    "--workdrive",
    type=str,
    prompt="Specify the workdrive",
    callback=validate_work_drive,
)
@click.option("--depot", type=str, prompt="Specify the depot")
def create_client(workdrive, depot):
    """create a new client for the given depot"""

    do_create_client(workdrive, depot)


def do_create_client(workdrive, depot, stream="main"):
    # Our standard clients are rooted at the P drive directly.

    client_name = get_client_name(depot)
    client_root = os.path.normpath(os.path.join(tools.subst_drive_get, "/", depot))
    date = datetime.datetime.now(tz=datetime.timezone.utc).strftime(
        "%Y/%m/%d %H:%M:%S %Z"
    )
    options = "rmdir"
    if depot == "UE":
        options += " allwrite"

    client_spec = f"""
Client: {client_name}
Owner: {get_username()}
Host: {get_hostname()}
Description:
 Created by paxo on {date}
Root: {client_root}
Options: {options}
Stream: //{depot}/{stream}
View:
 //{depot}/{stream}/... //{client_name}/...
"""
    print(client_spec)
    os.makedirs(client_root, exist_ok=True)
    p4_run(["client", "-i"], input=client_spec.encode(), check=True)
    write_p4config(client_name, client_root)
    return client_name


def write_p4config(client_name, client_root):
    p4config = f"""
P4CLIENT={client_name}
"""
    with open(os.path.join(client_root, ".p4config"), "w") as f:
        f.write(p4config)


def get_client_name(depot, postfix=None):
    """return a suitable client name for the given stream"""
    name = f"pd.{get_username()}.{get_hostname()}.{depot}"
    if postfix:
        name += f".{postfix}"
    return name


def get_hostname():
    return socket.gethostname()


def get_username():
    # we want to use the same username as perforce uses in p4user
    # but if that is not set, use the current user
    global p4_user
    if p4_user:
        return p4_user
    user = p4_set_get()["P4USER"]
    if not user:
        user = getpass.getuser()
    p4_user = user
    return user


def p4client_set(client):
    """set the current client"""
    p4_run(["set", f"P4CLIENT={client}"], check=True)


def p4client_get():
    """get the current client"""
    out = p4_run(["set", "P4CLIENT"], check=True, capture_output=True)
    return out.stdout.decode().strip().split("=")[1].split()[0]


@contextlib.contextmanager
def p4client(client):
    """context manager for setting the current client"""
    old_client = p4client_get()
    p4client_set(client)
    try:
        yield
    finally:
        p4client_set(old_client)


def p4_sync(depot="paxdei", all=False):
    if all or depot == "all":
        depots = ["paxdei", "UE"]
    else:
        depots = [depot]
    for d in depots:
        client = get_client_name(depot=d)
        with p4client(client):
            p4_run(["sync"], check=True)


@p4.command()
@click.option("--depot", type=str, default="paxdei")
@click.argument("stream-name")
def switch(depot, stream_name):
    """switch to the given stream"""
    depots = ["paxdei", "UE"]
    use_fallback = "paxdei" in depots
    fallback = "main"
    rollback = []
    fail = False
    try:
        for d in depots:
            # fallback to main if the stream doesn't exist and we are not paxdei
            current_fallback = fallback if use_fallback and d != "paxdei" else None
            previous_stream = get_client_stream(get_client_name(d))
            result = p4_switch(d, stream_name, fallback=current_fallback)
            if result:
                rollback.append((stream_name, previous_stream))
            else:
                fail = True
                break
        else:
            # we succeeded in switching all depots
            rollback = []
    finally:
        # rollback if we failed
        for depot, previous_stream in reversed(rollback):
            p4_switch(depot, previous_stream)

    return not fail


def p4_switch(depot, stream_name, fallback=None, no_sync=True):
    """switch to the given stream"""

    streams = find_streams(depot)
    if stream_name not in streams:
        if fallback and fallback in streams:
            print(
                f"stream {stream_name} not found for depot {depot}.  Using {fallback} instead"
            )
            stream_name = fallback
        else:
            print(f"stream {stream_name} not found for depot {depot}.")
            return False

    client = get_client_name(depot)
    current_stream = get_client_stream(client)
    print(depot, stream_name, "1")
    if current_stream == stream_name:
        print(f"client {client} is already on stream '{stream_name}'")
        return True

    cmd = ["switch", stream_name]
    if no_sync:
        cmd.insert(1, "--no-sync")
        print(cmd)
    res = p4_run(cmd, check=False, env={"P4CLIENT": client})
    print(depot, stream_name, "2")
    if res.returncode == 0:
        print(f"client {client} switched to stream '{stream_name}'")
        return True
    return res.returncode != 0


@p4.command()
@click.option("--depot", type=str, default="paxdei")
def stream(depot):
    """display the current stream"""

    client = get_client_name(depot)
    current_stream = get_client_stream(client)
    print(f"client {client} is on stream '{current_stream}'")


def get_client_info(client):
    """return the client info for the given client"""
    out = p4_json(["client", "-o", client])
    return out[0]


def get_client_stream(client):
    """return the stream for the given client"""
    out = p4_json(["switch"], env={"P4CLIENT": client})
    return out[0]["data"]


def fix_p4v_settings():
    # fix the p4v xml settings file
    filename = os.path.join(
        os.environ["USERPROFILE"], ".p4qt", "ApplicationSettings.xml"
    )

    if os.path.isfile(filename):
        tree = ElementTree.parse(filename)
        root = tree.getroot()
        for elem in root.iter():
            if (
                elem.tag == "String"
                and elem.attrib.get("varName", "") == "DefaultCharset"
            ):
                elem.text = "utf8-bom"
        tree.write(filename, encoding="utf-8")
    else:
        # create new settings with only this content.
        root = ElementTree.Element(
            "PropertyList",
            attrib={"IsManaged": "TRUE", "varName": "ApplicationSettings"},
        )
        elem = ElementTree.SubElement(
            root, "PropertyList", attrib={"IsManaged": "TRUE", "varName": "Connection"}
        )
        elem = ElementTree.SubElement(
            elem, "String", attrib={"varName": "DefaultCharset"}
        )
        elem.text = "utf8-bom"
        tree = ElementTree.ElementTree(root)
        ElementTree.indent(tree, " ")
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        tree.write(filename, encoding="utf-8")


@p4.command()
@click.option("--depot", type=str, default="paxdei")
def streams(depot):
    """list the streams in the given depot"""
    streams = find_streams(depot)
    for stream in streams:
        print(stream)


def find_streams(depot):
    """return a list of streams in the given depot"""
    client = get_client_name(depot)
    out = p4_json(["switch", "-l"], env={"P4CLIENT": client})
    return [line["data"].split()[0] for line in out]


def get_engine_path():
    """Return the root path of the engine"""
    client = get_client_name(depot="UE")
    client_info = get_client_info(client)
    return client_info["Root"]



import glob
import os.path
import re
import stat
import uuid
from typing import List, Optional, Tuple

from .desktop_win import (
    enumerate_engine_installations,
    normalize_root,
    register_engine_installation,
)

# stuff which helps with the engine management on the desktop. Based on the
# DesktopPlatformBase class from the UnrealEngine.


def is_guid(s: str) -> bool:
    """Check if the string is a guid"""
    try:
        uuid.UUID(s)
        return True
    except ValueError:
        return False


def get_engine_descriptions() -> List[Tuple[str, str, str]]:
    """Get the descriptions of the engines installed on the system."""
    for engine_id, root_dir in enumerate_engine_installations().items():
        yield (engine_id, root_dir, get_engine_description(engine_id, root_dir))


def get_engine_description(identifier: str, root_dir=None) -> str:
    """Get the description of the engine from the identifier."""
    if is_stock_engine_release(identifier):
        return identifier

    # Otherwise get the path
    root_dir = root_dir or get_engine_root_dir_from_identifier(identifier)
    if not root_dir:
        return ""

    platform_root_dir = os.path.normpath(root_dir)

    # source build
    if is_source_distribution(root_dir):
        return f"Source build at {platform_root_dir}"
    else:
        return f"Binary build at {platform_root_dir}"


def is_stock_engine_release(identifier: str) -> bool:
    """Check if the identifier is a stock engine release."""
    try:
        uuid.UUID(identifier)
        return False
    except ValueError:
        pass

    # classic UE uses only uuids for non-stock engines.  But we allow for
    # other engines, so only if this is a x.y version number do we consider it stock
    match = re.match(r"(\d+\.\d+)", identifier)
    return bool(match)


def get_engine_root_dir_from_identifier(engine_id: str) -> Optional[str]:
    """Look up the engine path from the registry"""
    engines = enumerate_engine_installations()
    return engines.get(engine_id, None)


def get_engine_identifier_from_root_dir(root_dir: str) -> str:
    root_dir = normalize_root
    engines = enumerate_engine_installations()

    # we prefer GUID keys to custom keys.
    found = None
    for k, v in engines.items():
        if v == root_dir:
            if is_guid(k):
                return k
            if not found:
                found = k
    if found:
        return found

    # otherwise, just add it
    return register_engine_installation(root_dir)


def is_valid_root_directory(engine_root):
    """See if this is a proper engine root"""

    # 1 there needs to be an Engine/Binaries folder
    if not os.path.isdir(os.path.join(engine_root, "Engine", "Binaries")):
        return False
    # 2 Also check there's an Engine\Build directory.
    # This will filter out anything that has an engine-like directory structure
    # but doesn't allow building code projects - like the launcher.
    if not os.path.isdir(os.path.join(engine_root, "Engine", "Build")):
        return False

    # 3 Check for a Build.version file.  This will rule out empty directory structures
    if not os.path.isfile(
        os.path.join(engine_root, "Engine", "Build", "Build.version")
    ):
        return False

    # else, we are ok
    return True


def is_source_distribution(engine_root):
    """Check if this is a source distribution"""
    filename = os.path.join(engine_root, "Engine", "Build", "SourceDistribution.txt")
    return os.path.isfile(filename) and stat.stat(filename).st_size > 0


def is_perforce_build(engine_root):
    """Check if this is a perforce build"""
    filename = os.path.join(engine_root, "Engine", "Build", "PerforceBuild.txt")
    return os.path.isfile(filename) and stat.stat(filename).st_size > 0


class ProjectDirectory:
    cache = {}

    def __init__(self, root):
        self.root = root
        # project dictionary for a given engine
        roots, projects = self.get_project_dictionary(root)
        self.project_root_dirs = roots
        self.projects = projects

    @classmethod
    def get_project_dictionary(cls, engine_root):
        roots = []
        engine_root = os.path.abspath(engine_root)
        for file in glob.glob(os.path.join(engine_root, "*.uprojectdirs")):
            with open(file) as f:
                for line in f:
                    line = line.strip()
                    if not line or line.startswith(";"):
                        continue

                    # convert to absolute path
                    if not os.path.isabs(line):
                        line = os.path.join(engine_root, line)
                    line = os.path.normcase(os.path.abspath(line))

                    # is it under the root
                    if line.startswith(os.path.normcase(engine_root)):
                        roots.append(line)
                    else:
                        print(
                            f"Warning: project search path {line} is not under engine root"
                        )

        # Search for all the projects under each root directory
        projects = []
        for root in roots:
            uprojects = glob.glob(os.path.join(root, "*", "*.uproject"))
            for up in uprojects:
                projects.append(
                    (os.path.basename(up), os.path.normcase(os.path.abspath(up)))
                )
        return roots, projects

    @classmethod
    def get(cls, engine_root):
        engine_root = os.path.abspath(engine_root)
        if engine_root in cls.cache:
            return cls.cache[engine_root]

        # create a new one
        result = cls(engine_root)
        cls.cache[engine_root] = result
        return result

    def is_foreign_project(self, project_file_name):
        # a foreign project is one which is not 'native', i.e. outside the engine root
        # check if this project is within the engine
        project_file_name = os.path.normcase(os.path.abspath(project_file_name))
        projects = [x[1] for x in self.projects]
        if project_file_name in projects:
            return False
        # could be a new project, check if its parent dir is in the project root dirs
        project_root_dir = os.path.dirname(project_file_name)
        if project_root_dir in self.project_root_dirs:
            return False

        # ok, must be foreign, then
        return True
